#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Python client for the Allgo API.

This module provides a class that simplify the use of the Allgo API
by dealing with the requests library.
Basically you will create a Client object, then launch all the jobs you want,
get their status, and be able to download the generated files.

The module provides:
 constantes
   | MAIN_INSTANCE_URL - The default Allgo url used.
   | TOKEN_FILE - The file containing the user token.

 class
   | Client - The main class to use the API easily.

 Errors
   | AllgoError  - generic module error.
   | TokenError  - Token related error.
   | StatusError - API connexion related error.

Example :
  .. code-block:: python

    import allgo

    client = allgo.Client()

    out_dict1 = client.create_job(the_app, params=the_params, files=the_files)
    out_dict2 = client.job_status(out_dict1['id'])

or even quicker :
  .. code-block:: python

    out_dict2 = client.run_job(the_app, params=the_params, files=the_files)

and finally :
  .. code-block:: python

    url = out_dict[str(job_id)]['allgo.log']
    filepath  = client.download_file(url)
"""

# standard libs
from __future__ import print_function
from distutils.version import StrictVersion
import errno
import logging
import os
import sys

# from json.decoder import JSONDecodeError  # python3
from os.path import expanduser

import requests

__version__ = StrictVersion('1.1.0')

#: The default Allgo url used.
MAIN_INSTANCE_URL = "https://allgo18.inria.fr"

#: The file containing the user token.
#: It is read in last resort to get a token.
TOKEN_FILE = os.path.join(expanduser("~"), '.allgo_token')


# ===============================================================
class AllgoError(Exception):
    """Generic error generated by this module."""


class TokenError(AllgoError):
    """Raise TokenError when none is set."""

    def __init__(self):
        err_msg  = "You must provide a token in parameter"
        err_msg += " or define an environment variable 'ALLGO_TOKEN'"
        err_msg += " or write your token in {}".format(TOKEN_FILE)
        super(TokenError, self).__init__(err_msg)


class StatusError(AllgoError):
    """Raise StatusError when API does not return a 200 code.

    .. code-block:: python

       try:
           something(client_object)
       except StatusError as e:
           print("status: {}, msg: {}".format(e.status_code, e.msg))
    """

    def __init__(self, status_code, response, *args, **kwargs):
        """TODO"""
        self._status_code = status_code
        self._response = response
        AllgoError.__init__(self, *args, **kwargs)

    @property
    def status_code(self):
        return self._status_code

    @property
    def msg(self):
        return self._response.json()['error']


def _local_token():  # pragma: no cover
    """read and return the content of TOKEN_FILE or None."""
    if os.path.exists(TOKEN_FILE):
        with open(TOKEN_FILE) as f:
            return f.read()
    return None


# ---------------------------------------------------------------
class Client:
    """
    manage connexions to the API and provide methods to deal with it.

    One object manage one user-Allgo connexion.

     accessors
       | allgo_url - the url of the instance the user want to ask to.
       | token     - the user token linked to the instance. cf {allgo_url}/profile
       | verify_tls- does tls have to be verified ?

     methods
       | app_metrics - retrieve the metrics related to an app.
       | create_job  - create an app's job.
       | download_file - download a file generated by a job.
       | job_status  - get status of a job.
       | run_job     - short cut combine create_job+loop on job_status until the job is done.
    """

    def __init__(self, token=None, allgo_url=MAIN_INSTANCE_URL, verify_tls=True):
        """Build an object to manage the parameters of a user-connexion to an Allgo-instance.

        Parameters
        ----------
        token : string, optional
            the user's token for the querying Allgo instance.
            if not provided, we check ALLGO_TOKEN env variable and TOKEN_FILE
        allgo_url  : url, optional [=MAIN_INSTANCE_URL]
            url of the Allgo instance to query.
        verify_tls : bool, optional [=True]
            set to False if you don't want to force TLS verification (not recommanded).

        Raises
        ------
        TokenError
            if no token can be determined.
        """
        self._allgo_url  = allgo_url
        self._verify_tls = verify_tls
        self._token = token or os.getenv('ALLGO_TOKEN') or _local_token()
        if None is self._token:
            raise TokenError()

    @property
    def token(self):
        """return the user token for the querying all instance."""
        return self._token

    @property
    def allgo_url(self):
        """return the url of the querying allgo instance."""
        return self._allgo_url

    @property
    def verify_tls(self):
        """return the _verify_tls attribute."""
        return self._verify_tls

    def _request_api(self, method, url_end, **kwargs):
        """Build and launch a http request on the Allgo instance API.

        Parameters
        ----------
        method : string
            'post' or 'get', the method to use in the http request.
        url_end : string
            end of the url to place after the API "signature".
        kwargs : dict
            any named parameters to use in the API request.

        Returns
        -------
        dictionnary or Response
            the API json response as a dictionnary if possible.
            if not (like for downloading files), the requests Response object.

        Raises
        ------
        StatusError
            with the Response status code if it's not 200.
        AllgoError
            if the method is unknown.
        """
        headers = {'Authorization': 'Token token={}'.format(self.token)}
        url = '{}/api/v1/{}'.format(self.allgo_url, url_end)
        logging.debug("querying '%s' with method '%s'", str(url), str(method))
        logging.debug("and with data = %s", str(kwargs))

        actions = {
            'post': requests.post,
            'get' : requests.get,
        }
        try:
            resp = actions[method](url, headers=headers, verify=self.verify_tls, **kwargs)
            logging.debug("raw API answer : %s", str(resp.content))
            if resp.status_code != requests.codes.ok:
                logging.debug("req.status_code = %s", str(resp.status_code))
                raise StatusError(resp.status_code, resp)
            return resp.json()
        except ValueError:  # pure python3 should use JSONDecodeError
            # this happens when response content cannot be converted to json
            #   for example when downloading a file.
            return resp
        except KeyError as err:
            msg  = "method {} is not allowed to request the API.".format(method)
            msg += "KeyError: {}".format(str(err))
            raise AllgoError(msg)

    def create_job(self, app, version=None, params='', files=None, queue=None):
        """Create a job for app, with params.

        Launch the API request to create a job for the given app,
        with parameters provided.

        Parameters
        ----------
        app : string or int
            application name or id.
        version: string, optional
            version of the app to use.
            **Not used yet.**
        params : string, optional
            all the app command line params as one string.
        files  : list of filepaths, optional
            filepaths to upload.
        queue  : string
            name of the queue to use.

        Returns
        -------
        dictionnary
            the API output as a dictionnary.
            keys 'id' and 'url'.

        Raises
        ------
        StatusError
            with the Response status code if it's not 200.
        """
        data = {
            # ~ "job[webapp_name]": app,
            "job[webapp_id]"  : app,
            "job[param]"      : params,
            "job[version]"    : version,
            "job[queue]"      : queue,
        }

        file_dict = {}
        if None is files:
            files = []
        for i, file_ in enumerate(files):
            key = "file[{}]".format(i)
            file_dict[key] = open(file_, 'rb')

        return self._request_api('post', url_end='jobs', data=data, files=file_dict)

    def job_status(self, job_id):
        """Get the status of a job as a dict.

        Parameters
        ----------
        job_id : int/string
            the job id to query.

        Returns
        -------
        dictionnary
            the API output as a dictionnary.
            keys 'status' and the job_id which value is a dictionnary filename: url.

        Raises
        ------
        StatusError
            with the Response status code if it's not 200.
        """
        url = 'jobs/{}'.format(job_id)
        return self._request_api('get', url)

    def app_metrics(self, app_id, what, from_=None, to_=None, step=None):
        """query the instance for some metrics.

        Owner of an app can query the instance to get some metrics.
        metrics are the number of jobs over a period of time,
        from 'from_' to 'to_', with data aggregated by 'step'
        3 kind of metrics are proposed ('what' param), they are an aggregate condition.

        Parameters
        ----------
        app : int
            application id.
        what: { "per_user", "per_result", "created" }
            the kind of metrics we want.
            "per_user"   : to get the # of job grouped by user.
            "per_result" : to get the # of job grouped by result.
            "created"    : to get the # of job grouped by date a creation.
        from_ : date
            date of begining in format (yyyy-mm-dd).
        to_   : date
            date of end in format (yyyy-mm-dd).
        step  : { "year", "month", "day" }, optional
            step period used to aggregate data.
            for more info cf. https://docs.djangoproject.com/en/1.11/ref/models/database-functions/#trunc

        Returns
        -------
        dictionnary
          .. code-block:: python

            {
              the app name:
                {
                    'data': [
                        {
                            Record format is specific to each 'what'
                            cf https://allgo.gitlabpages.inria.fr/doc/api.html#output-format
                        },
                        ...
                    ],
                    'from': __fmt_date(from_),
                    'to'  : __fmt_date(to),
                    'step'  : step,
                }
            }
        """
        url = 'metrics/{}/{}'.format(what, app_id)
        params = {
            'from' : from_,
            'to'   : to_,
            'step' : step,
        }
        return self._request_api('get', url, params=params)

    # --------------------------------------------
    # -- Accessory methods, to ease user work.
    # --------------------------------------------
    def _request_file(self, file_url):
        """Build and launch a http request on any URL, but with Allgo authentication.

        The file downloads as part of the API, it is an additionnal service.

        Parameters
        ----------
        file_url : string/URL
            the URL to download the file from.

        Returns
        -------
        filepath
            the local filepath of the file downloaded.

        Raises
        ------
        StatusError
            with the Response status code if it's not 200.
        """
        headers = {'Authorization': 'Token token={}'.format(self.token)}
        resp = requests.get(file_url, headers=headers, verify=self.verify_tls)

        if resp.status_code != requests.codes.ok:
            raise StatusError(resp.status_code, resp)
        return resp

    def download_file(self, file_url, outdir='.', force=False):
        """download the file pointed by the url.

        The file is downloaded in the outdir directory (which is created if not exist).
        The old Allgo site (allgo.inria.fr) doesn't need authentication.
        The new allgo site (allgo18.inria.fr) does need authentication.

        Parameters
        ----------
        file_url : string/URL
            the URL to download the file from.
        outdir : dirpath, optional [='.']
            the dirpath to write file to.
        force : bool, optional [=False]
            set to true to force erasing existing file.

        Returns
        -------
        filepath
            the local filepath of the file downloaded.

        Raises
        ------
        FileExistsError
            if the output file exists and force is false.
        StatusError
            with the Response status code if it's not 200.
        """
        if 2 == sys.version_info.major:
            return self.__download_file_p2(file_url, outdir, force)
        return self.__download_file_p3(file_url, outdir, force)

    def __download_file_p3(self, file_url, outdir='.', force=False):
        mode = 'wb' if force else 'xb'

        resp = self._request_file(file_url)

        # urlparse is better, but a pain to use in python2 & 3
        #   fileurl, if come from job_status doesn't have params, so it's ok.
        filename = os.path.basename(file_url)
        os.makedirs(outdir, exist_ok=True)
        outfilepath = os.path.join(outdir, filename)

        with open(outfilepath, mode) as file_out:
            # Note: '8192' = 2^13 : size of cache/buffer.
            #   = max size transfered in one operation.
            for chunk in resp.iter_content(chunk_size=8192):
                if chunk:
                    file_out.write(chunk)

        return outfilepath

    def __download_file_p2(self, file_url, outdir='.', force=False):
        # python2: outdir cannot be a PosixPath (cf makedirs, os.path.join)
        outdir = str(outdir)

        filename = os.path.basename(file_url)
        outfilepath = os.path.join(outdir, filename)

        # method with os.fdopen(os.open(outfilepath, mode), 'w') doesn't exist
        if os.path.isfile(outfilepath) and not force:
            msg = "file {} exists.".format(outfilepath)
            raise OSError(errno.EEXIST, msg)

        try:
            os.makedirs(outdir)
        except OSError as e:
            # silence directory exists error
            if e.errno != errno.EEXIST:
                raise

        resp = self._request_file(file_url)
        with open(outfilepath, 'wb') as file_out:
            # Note: '8192' = 2^13 : size of cache/buffer.
            #   = max size transfered in one operation.
            for chunk in resp.iter_content(chunk_size=8192):
                if chunk:
                    file_out.write(chunk)

        return outfilepath

    def run_job(self, app, version=None, params='', files=None,
                sleep_duration=2, verbose=False):
        """Create a job and wait for it to terminate.

            Simple combinaison of create_job followed by several call to job_status
            until the job is done.

        Parameters
        ----------
        app : string or int
            application name or id.
        version: string, optional
            version of the app to use.
            **Not used yet.**
        params : string, optional
            all the app command line params as one string.
        files  : list of filepaths, optional
            filepaths to upload.
        sleep_duration : int, optional [=2]
            number of seconds between 2 job_status queries.
        verbose : bool, optional [=false]
            if true, will print a message to show progression.
            a dot will be printed each time a job_status query is made (like a progress bar),
            so a dot each sleep_duration.

        Returns
        -------
        dictionnary
            the API output as a dictionnary.
            keys 'status' and the job_id which value is a dictionnary filename: url.

        Raises
        ------
        StatusError
            with the Response status code if it's not 200.
        """
        from time import sleep
        # Note: states new, and deleted are almost impossible to get.
        STATES = ['NEW', 'WAITING', 'RUNNING', 'ARCHIVED', 'DONE', 'DELETED', 'ABORTING']

        out_dict = self.create_job(app, version=version, params=params, files=files)
        job_id = out_dict['id']
        flags = { s.lower(): False for s in STATES }

        while True:
            sleep(sleep_duration)

            out_dict = self.job_status(job_id)
            status = out_dict['status']

            if True is verbose:
                if False is flags[status]:
                    print("\n{}\t".format(status), end='')
                    flags[status] = True
                else:
                    print(".", end='')
                sys.stdout.flush()

            if status == 'done':
                if True is verbose:
                    print("")  # to get a return line
                out_dict['id'] = job_id
                return out_dict

        # should never goes here

    def get_log(self, job_id):
        """Generator of log file content.

        Parameters
        ----------
        job_id : int/string
            the job id to query.

        Returns
        -------
        string generator
            The object produces a sequence of allgo.log content.

        Raises
        ------
        AllgoError
            if the job is not done.
        StatusError
            with the Response status code if it's not 200.
        """
        ALLGO_LOG = 'allgo.log'

        # we have to wait for the end of the job.
        out_dict = self.job_status(job_id)
        if 'done' != out_dict['status']:
            raise AllgoError("The job is not done, you cannot access the log yet.")

        # job is done, get the log file
        # copy code of _request_api method with alteration
        file_url = out_dict[str(job_id)][ALLGO_LOG]
        resp = self._request_file(file_url)

        for chunk in resp.iter_content(chunk_size=8192, decode_unicode=True):
            # the decode function is needed to turn byte to string
            yield chunk.decode()
